/*

	GJ Java Puzzle.



	Copyright (c) 1996, DMC K Morikawa

	

*/



import java.awt.*;

import java.awt.image.*;

import java.awt.FontMetrics;

import java.applet.*;

import java.net.URL;

import java.net.MalformedURLException;

import java.lang.Math;









public class Puzzle extends Applet implements Runnable

{

	String title         = null;  // Title of the puzzle

	String result        = null;  // Message to display when solved.

	String imageSource   = null;  // Base of the names of the image files.

	int numRows          = 0;     // Number of rows in the puzzle.

	int numColumns       = 0;     // Number of columns in the puzzle.

	int numPieces        = 0;     // Rows * Columns

	int blankNumber      = 0;     // Which piece is the blank piece?

	int imageSizeX       = 0;     // Physical size of each image (Pixels across)

	int imageSizeY       = 0;     // Physical size of each image (Pixels down)

	int pieceSizeX       = 0;     // No of pixels to use on the screen (across)

	int pieceSizeY       = 0;     // No of pixels to use on the screen (down)

	int thisPiece        = 0;     // Used for seeing if a piece can be moved.

	int thatPiece        = 0;     //   ditto.

	int DescrambleThreshold  = 200;  // Number of moves before descrambling is allowed.



	Rectangle mat        = new Rectangle(0,0,0,0);

	int frameWidth       = 3;     // Width of the frame around each piece.



	boolean initError    = false; // Error during initialisation?

	int initErrorCode    = 0;     // Code of any error encountered.

	boolean runError     = false; // Error during running of the puzzle?



	boolean debug        = true;  // Are we in debug mode - Console output.



	boolean initialPaint = true;  // Is this first time a repaint() is called?

	Image piece[];		      // Images of each piece of the puzzle.

	int   slot[];		      // Which piece is in which position in the puzzle.

	int	 numMoves     = 0;    // Number of moves by the user.





   

   void dbg(String s)

   {  if (debug)

	 System.out.println(s);

      else

	 showStatus(s);

   }





   

   boolean puzzleSolved()

   {  int p;



      for (p=0; p<numPieces; p++)

	 if (slot[p] != p) return(false);



      return(true);

   }





   

   int canMove(int row, int column)

   {

	thisPiece = row*numColumns + column;



	

	if (column>0)

	{  thatPiece = row*numColumns+(column-1);

	      if (thatPiece < numPieces)

		 if (slot[thatPiece] == blankNumber)

			 return(1);

	}





	if (column<(numColumns-1))

	{  thatPiece = row*numColumns+(column+1);

	      if (thatPiece < numPieces)

		 if (slot[thatPiece] == blankNumber)

		    return(1);

	}



	

	if (row>0)

	{  thatPiece = (row-1)*numColumns+column;

	      if (thatPiece < numPieces)

		 if (slot[thatPiece] == blankNumber)

		    return(1);

	}



	

	if (row<(numRows-1))

	{  thatPiece = (row+1)*numColumns+column;

	      if (thatPiece < numPieces)

		 if (slot[thatPiece] == blankNumber)

		    return(1);

	}



	return(0);

   }





   

   void solvePuzzle()

   {  for (int i=0; i<numPieces; i++)

	 slot[i] = i;

   }





   

   void scramblePieces()

   {

      int p, p1, temp, row, column;

      int randMoves;

      boolean moveNotValid;



      // Start puzzle in default, 'solved' position.

      solvePuzzle();



      randMoves = 101 + (int) (Math.random()*50.00);



      

      for (int i=0; i<randMoves; i++)

      {  moveNotValid=true;

	 showStatus("Loading : Rand Move # " + i + " of " + randMoves);



	 do 

	 {

	    p1 = (int) (Math.random() * 65535.00);



	    p = p1 % numPieces;

	    row = p / numColumns;

	    column = p % numColumns;



	    if (canMove(row, column) == 1)

	       moveNotValid = false;



	 } while (moveNotValid);



	 temp = slot[thisPiece];

	 slot[thisPiece] = slot[thatPiece];

	 slot[thatPiece] = temp;

      }



      showStatus("");

   }





   

   public void init()

   {



     

      showStatus("Loading : Decoding parameters");



      // OPTIONAL Parameter

      showStatus("Loading : title ...");

      String param = getParameter("title");	dbg("title = "+param);

      if (param != null)

	 title = new String(param);

      else

	 title = "GJ Java Puzzle";



      // OPTIONAL Parameter

      showStatus("Loading : result ...");

      param = getParameter("result");		dbg("result = "+param);

      if (param != null)

	 result = new String(param);

      else

	 result = "The puzzle has been solved.";



      // MANDATORY Parameter

      showStatus("Loading : imagesource ...");

      param = getParameter("imagesource");	dbg("imagesource = "+param);

      if (param != null)

	 imageSource = new String(param);

      else

      {  initErrorCode = 1;

	 initError = true;

      }



      // OPTIONAL Parameter

      showStatus("Loading : descramble from ...");

      param = getParameter("descramblefrom");	dbg("descramble from = "+param);

      if (param != null)

	 DescrambleThreshold = Integer.parseInt(param);



      // MANDATORY Parameter

      showStatus("Loading : rows ...");

      param = getParameter("rows");		dbg("rows = "+param);

      if (param != null)

	 numRows = Integer.parseInt(param);

      else

      {  initErrorCode = 1;

	 initError = true;

      }



      // MANDATORY Parameter

      showStatus("Loading : columns ...");

      param = getParameter("columns");		dbg("columns = "+param);

      if (param != null)

	 numColumns = Integer.parseInt(param);

      else

      {  initErrorCode = 2;

	 initError = true;

      }



      // MANDATORY Parameter

      showStatus("Loading : xsize ...");

      param = getParameter("xsize");          dbg("xsize = "+param);

      if (param != null)

	 imageSizeX = Integer.parseInt(param);

      else

      {  initErrorCode = 3;

	 initError = true;

      }



      // MANDATORY Parameter

      showStatus("Loading : ysize ...");

      param = getParameter("ysize");		dbg("ysize = "+param);

      if (param != null)

	 imageSizeY = Integer.parseInt(param);

      else

      {  initErrorCode = 4;

	 initError = true;

      }



      // OPTIONAL Parameter

      showStatus("Loading : xpixels ...");

      param = getParameter("xpixels");          dbg("xpixels = "+param);

      if (param != null)

	 pieceSizeX = Integer.parseInt(param);

      else

	 pieceSizeX = imageSizeX;



      // OPTIONAL Parameter

      showStatus("Loading : ypixels ...");

      param = getParameter("ypixels");		dbg("ypixels = "+param);

      if (param != null)

	 pieceSizeY = Integer.parseInt(param);

      else

	 pieceSizeY = imageSizeY;



      // MANDATORY Parameter

      showStatus("Loading : blank ...");

      param = getParameter("blank");         	dbg("blank = "+param);

      if (param != null)

	 blankNumber = Integer.parseInt(param) - 1;

      else

      {  initErrorCode = 5;

	 initError = true;

      }



      if (!initError)

      {

	 

	 int w, h;



	 w = pieceSizeX * numColumns + (numColumns+1) * frameWidth;

	 h = pieceSizeY * numRows    + (numRows+1)    * frameWidth;

	 mat.resize(w, h);



	 // Identify how many pieces are in this puzzle.

	 numPieces = numRows * numColumns;



	 // Initialise randomly located puzzle pieces.

	 showStatus("Loading : slots ...");

	 slot  = new int[numPieces];

	 scramblePieces();



	 // Initialise and load pieces.

	 showStatus("Loading : puzzle ...");

	 piece = new Image[numPieces];

	 for (int p = 1 ; p <= numPieces; p++)

	    piece[p-1] = getImage(getCodeBase(), imageSource + p + ".jpg");

      }

      thisPiece = thatPiece = -1;

   }





 

   public void start()

   {  if (initError == true)

	 showStatus("Init Error : Applet can not run. Error " + initErrorCode);

      else

      {

	 showStatus("Puzzle : " + title + " running.");

      }

   }





   

   public void stop()

   {

   }





   

   public boolean handleEvent(Event evt)

   {

      if (evt.id == Event.MOUSE_DOWN)

      {

	 int row = -1, column = -1, pieceNumber = -1;



	 // Identify piece that is to be tested for moveability.

	 column = evt.x / (pieceSizeX+frameWidth);

	 if ((evt.x % (pieceSizeX+frameWidth)) < frameWidth)

	    column = -1;

	 if (column >= numColumns) column = -1;



	 row = evt.y / (pieceSizeY+frameWidth);

	 if ((evt.y % (pieceSizeY+frameWidth)) < frameWidth)

		 row = -1;

	 if (row >= numRows) row = -1;



	 // Can this piece be moved?

	 if (row != -1 && column != -1)

	 {  pieceNumber = row * numColumns + column;

	    if (slot[pieceNumber] == blankNumber && numMoves >= DescrambleThreshold)

	    {  solvePuzzle();

	       repaint();

	    }

	    else if (canMove(row, column) == 1)

	    {  int temp;

	       temp = slot[thisPiece];

	       slot[thisPiece] = slot[thatPiece];

	       slot[thatPiece] = temp;



	       numMoves++;

	       // play(getCodeBase(), "move.au"); <-- Was annoying...



	       row = thisPiece / numColumns;

	       column = thisPiece % numColumns;

	       repaint(mat.x+(pieceSizeX+frameWidth)*column+frameWidth,

		       mat.y+(pieceSizeY+frameWidth)*row+frameWidth,

		       pieceSizeX, pieceSizeY);



	       row = thatPiece / numColumns;

	       column = thatPiece % numColumns;

	       repaint(mat.x+(pieceSizeX+frameWidth)*column+frameWidth,

		       mat.y+(pieceSizeY+frameWidth)*row+frameWidth,

		       pieceSizeX, pieceSizeY);

	    }

	 }

	 return true;

      }

      else

	 return super.handleEvent(evt);

    }



    

    public void run()

    {

	repaint();

      /*

	while (true)

	{

	//repaint();



	}

      */

    }



    

    public void paint(Graphics g)

    {

	if (initError || runError)

	    showStatus("Paint : Applet can not run. Error " + initErrorCode);

	else

	{  Color frameColor = Color.blue;

	   Color insetColor = Color.white;

	   Color textColor  = Color.black;



	   if (true)

	   {

	      if (initialPaint)

	      {  // Request each piece

		 for (int row=0, p=0; row<numRows; row++)

		    for (int column=0; column<numColumns; column++, p++)

		    {  showStatus("Loading puzzle... " + p);

		       prepareImage(piece[slot[p]], pieceSizeX, pieceSizeY, this);

		       g.drawImage(piece[slot[p]],

				   mat.x+(pieceSizeX+frameWidth)*column+frameWidth,

				   mat.y+(pieceSizeY+frameWidth)*row+frameWidth,

				   pieceSizeX, pieceSizeY,

				   this);



		    }



	      }

	      else

	      {

		 boolean haveAllImages=true;

		 int numImagesGot=0;



		 for (int p=0; p<numPieces; p++)

		    if ((checkImage(piece[slot[p]], pieceSizeX, pieceSizeY, this) & ALLBITS) == 0)

		       haveAllImages = false;

		    else

		       numImagesGot++;



		 if (haveAllImages)

		 {

		    // Select the colour of the bounding frame

		    g.setColor(frameColor);



		    // Draw to full puzzle bounds.

		    g.drawRect(mat.x, mat.y, mat.width-1, mat.height-1);



		    // Draw a bounding box around each of the piece squares

		    for (int row=0; row<numRows; row++)

		       for (int column=0; column<numColumns; column++)

		       {  int pieceX, pieceY;



			  pieceX = mat.x+column*(frameWidth+pieceSizeX)+frameWidth-1;

			  pieceY = mat.y+row*(frameWidth+pieceSizeY)+frameWidth-1;



			  g.drawRect(pieceX, pieceY,

				     pieceSizeX+frameWidth-2,

				     pieceSizeY+frameWidth-2);

		       }



		    // Select the colour of the inset frames

		    g.setColor(insetColor);



		    // Draw an inset box around each of the piece squares

		    for (int row=0; row<numRows; row++)

		       for (int column=0; column<numColumns; column++)

		       {  int pieceX, pieceY;



			  pieceX = mat.x+column*(frameWidth+pieceSizeX)+frameWidth-2;

			  pieceY = mat.y+row*(frameWidth+pieceSizeY)+frameWidth-2;



			  g.drawRect(pieceX, pieceY,

				     pieceSizeX+frameWidth,

				     pieceSizeY+frameWidth);

		       }



		    // Draw each piece

		    for (int row=0, p=0; row<numRows; row++)

		       for (int column=0; column<numColumns; column++, p++)

		       {  if (true)

			  {  boolean showBlank = true;

			     if (puzzleSolved())

			     {  showBlank = false;

				showStatus("Your Moves : " + numMoves + " : " + result);

			     }

									  else

			     showStatus("Your Moves : " + numMoves + " : " + title);



			     if (slot[p] == blankNumber && showBlank)

			     {

				g.clearRect(mat.x+(pieceSizeX+frameWidth)*column+frameWidth,

					    mat.y+(pieceSizeY+frameWidth)*row+frameWidth,

					    pieceSizeX, pieceSizeY);



				if (numMoves >= DescrambleThreshold)

				{  int hei, top, left;

				   String outText = null;

				   FontMetrics fm = g.getFontMetrics();



				   g.setColor(textColor);



				   hei = fm.getHeight();

					top = (pieceSizeY - 1*hei) / 2 + fm.getMaxAscent();



				   outText = "need help?";

				   left = (pieceSizeX - fm.stringWidth(outText)) / 2;

				   g.drawString(outText,

					     mat.x+(pieceSizeX+frameWidth)*column+frameWidth + left,

					     mat.y+(pieceSizeY+frameWidth)*row+frameWidth + top);

				}

			     }

			     else

			     {

				g.drawImage(piece[slot[p]],

					    mat.x+(pieceSizeX+frameWidth)*column+frameWidth,

					    mat.y+(pieceSizeY+frameWidth)*row+frameWidth,

					    pieceSizeX, pieceSizeY,

					    this);

			     }

			  }

		       }



		 } // If have all

		 else

		 {  // Text to say that we are loading right now...

		    // Select the colour of the text

		    showStatus("Now Loading Puzzle... " + numImagesGot + "/" + numPieces);

		 }



	      } // Not initial paint.

	   }

	}

	initialPaint = false;

    }

}

